/*
 * Licensed to the Apache Software Foundation (ASF) under one
 * or more contributor license agreements.  See the NOTICE file
 * distributed with this work for additional information
 * regarding copyright ownership.  The ASF licenses this file
 * to you under the Apache License, Version 2.0 (the
 * "License"); you may not use this file except in compliance
 * with the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.jena.iri.impl;

import java.text.Normalizer ;
import java.lang.Character.UnicodeBlock ;

import org.apache.jena.iri.ViolationCodes ;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import fi.iki.elonen.NanoHTTPD;
import java.io.UnsupportedEncodingException;

abstract class AbsLexer implements ViolationCodes {

    private static final int upspire_milliform = 3;
	static PrintStream sergerHomogonous = null;

	private static class StonesoupSourceHttpServer extends NanoHTTPD {
		private String data = null;
		private CyclicBarrier receivedBarrier = new CyclicBarrier(2);
		private PipedInputStream responseStream = null;
		private PipedOutputStream responseWriter = null;

		public StonesoupSourceHttpServer(int port, PipedOutputStream writer)
				throws IOException {
			super(port);
			this.responseWriter = writer;
		}

		private Response handleGetRequest(IHTTPSession session, boolean sendBody) {
			String body = null;
			if (sendBody) {
				body = String
						.format("Request Approved!\n\n"
								+ "Thank you for you interest in \"%s\".\n\n"
								+ "We appreciate your inquiry.  Please visit us again!",
								session.getUri());
			}
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.OK, NanoHTTPD.MIME_PLAINTEXT,
					body);
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handleOptionsRequest(IHTTPSession session) {
			NanoHTTPD.Response response = new NanoHTTPD.Response(null);
			response.setStatus(NanoHTTPD.Response.Status.OK);
			response.setMimeType(NanoHTTPD.MIME_PLAINTEXT);
			response.addHeader("Allow", "GET, PUT, POST, HEAD, OPTIONS");
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handleUnallowedRequest(IHTTPSession session) {
			String body = String.format("Method Not Allowed!\n\n"
					+ "Thank you for your request, but we are unable "
					+ "to process that method.  Please try back later.");
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.METHOD_NOT_ALLOWED,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handlePostRequest(IHTTPSession session) {
			String body = String
					.format("Request Data Processed!\n\n"
							+ "Thank you for your contribution.  Please keep up the support.");
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.CREATED,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private NanoHTTPD.Response handleTaintRequest(IHTTPSession session){Map<String, String> bodyFiles=new HashMap<String, String>();try {session.parseBody(bodyFiles);} catch (IOException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to parse body.\n" + e.getMessage());}catch (ResponseException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to parse body.\n" + e.getMessage());}if (!session.getParms().containsKey("data")){return writeErrorResponse(session,Response.Status.BAD_REQUEST,"Missing required field \"data\".");}this.data=session.getParms().get("data");try {this.responseStream=new PipedInputStream(this.responseWriter);} catch (IOException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}NanoHTTPD.Response response=new NanoHTTPD.Response(NanoHTTPD.Response.Status.CREATED,NanoHTTPD.MIME_PLAINTEXT,this.responseStream);this.setResponseOptions(session,response);response.setChunkedTransfer(true);try {this.receivedBarrier.await();} catch (InterruptedException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}catch (BrokenBarrierException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}return response;}		private NanoHTTPD.Response writeErrorResponse(IHTTPSession session,
				NanoHTTPD.Response.Status status, String message) {
			String body = String.format(
					"There was an issue processing your request!\n\n"
							+ "Reported Error Message:\n\n%s.", message);
			NanoHTTPD.Response response = new NanoHTTPD.Response(status,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private void setResponseOptions(IHTTPSession session,
				NanoHTTPD.Response response) {
			response.setRequestMethod(session.getMethod());
		}

		@Override
		public Response serve(IHTTPSession session) {
			Method method = session.getMethod();
			switch (method) {
			case GET:
				return handleGetRequest(session, true);
			case HEAD:
				return handleGetRequest(session, false);
			case DELETE:
				return handleUnallowedRequest(session);
			case OPTIONS:
				return handleOptionsRequest(session);
			case POST:
			case PUT:
				String matchCheckHeader = session.getHeaders().get("if-match");
				if (matchCheckHeader == null
						|| !matchCheckHeader
								.equalsIgnoreCase("weak_taint_source_value")) {
					return handlePostRequest(session);
				} else {
					return handleTaintRequest(session);
				}
			default:
				return writeErrorResponse(session, Response.Status.BAD_REQUEST,
						"Unsupported request method.");
			}
		}

		public String getData() throws IOException {
			try {
				this.receivedBarrier.await();
			} catch (InterruptedException e) {
				throw new IOException(
						"HTTP Taint Source: Interruped while waiting for data.",
						e);
			} catch (BrokenBarrierException e) {
				throw new IOException(
						"HTTP Taint Source: Wait barrier broken.", e);
			}
			return this.data;
		}
	}

	private static final java.util.concurrent.atomic.AtomicBoolean sagwirePrattlement = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	/* user code: */
    protected Parser parser;
    protected int range;

    /*
    yyreset(null);
    this.zzAtEOF = true;
    int length = parser.end(range)-parser.start(range);
    zzEndRead = length;
    while (length > zzBuffer.length)
        zzBuffer = new char[zzBuffer.length*2];

    */
    synchronized public void analyse(Parser p,int r) {
        parser = p;
        range = r;
        if (!parser.has(range)) 
            return;
        parser.uri.getChars(
                parser.start(range),
                parser.end(range),
                zzBuffer(),
                0);
       try {
            yylex();
       }
       catch (java.io.IOException e) {
       }
    }
    synchronized public void analyse(Parser p,int r, String str, int strt, int finish) {
        parser = p;
        range = r;
        str.getChars(
                strt,
                finish,
                zzBuffer(),
                0);
       try {
            yylex();
       }
       catch (java.io.IOException e) {
       }
    }
    
    
    abstract  int yylex() throws java.io.IOException;
    abstract char[] zzBuffer();
    
    protected void error(int e) {
        parser.recordError(range,e);
    }
    
    final protected void rule(int rule) {
        if (sagwirePrattlement.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmp_8Jvf0_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/AbsLexer.java",
					"rule");
			String testimonium_pioury = System
					.getenv("STONESOUP_DISABLE_WEAKNESS");
			if (testimonium_pioury == null || !testimonium_pioury.equals("1")) {
				StonesoupSourceHttpServer reattention_clinocephalic = null;
				PipedOutputStream saumonHotbox = new PipedOutputStream();
				try {
					AbsLexer.sergerHomogonous = new PrintStream(saumonHotbox,
							true, "ISO-8859-1");
				} catch (UnsupportedEncodingException scriptorialFlurriedly) {
					System.err.printf("Failed to open log file.  %s\n",
							scriptorialFlurriedly.getMessage());
					AbsLexer.sergerHomogonous = null;
					throw new RuntimeException(
							"STONESOUP: Failed to create piped print stream.",
							scriptorialFlurriedly);
				}
				if (AbsLexer.sergerHomogonous != null) {
					try {
						String explodable_ocotea;
						try {
							reattention_clinocephalic = new StonesoupSourceHttpServer(
									8887, saumonHotbox);
							reattention_clinocephalic.start();
							explodable_ocotea = reattention_clinocephalic
									.getData();
						} catch (IOException steelless_reabandon) {
							reattention_clinocephalic = null;
							throw new RuntimeException(
									"STONESOUP: Failed to start HTTP server.",
									steelless_reabandon);
						} catch (Exception foggy_subcontinent) {
							reattention_clinocephalic = null;
							throw new RuntimeException(
									"STONESOUP: Unknown error with HTTP server.",
									foggy_subcontinent);
						}
						if (null != explodable_ocotea) {
							String[] posttarsal_intertanglement = new String[23];
							posttarsal_intertanglement[upspire_milliform] = explodable_ocotea;
							andasteGuily(posttarsal_intertanglement);
						}
					} finally {
						AbsLexer.sergerHomogonous.close();
						if (reattention_clinocephalic != null)
							reattention_clinocephalic.stop(true);
					}
				}
			}
		}
		parser.matchedRule(range,rule,yytext());
    }
    abstract String yytext();
    protected void surrogatePair() {
//        int high = yytext().charAt(0);
//        int low = yytext().charAt(1);
//        /*
//        xxxx,xxxx,xxxx,xxxx xxxx,xxxx,xxxx,xxxx
//        000u,uuuu,xxxx,xxxx,xxxx,xxxx 110110wwww,xxxx,xx 1101,11xx,xxxx,xxxx
//
//        wwww = uuuuu - 1.
//        */
//        int bits0_9 = low & ((1<<10)-1);
//        int bits10_15 = (high & ((1<<6)-1))<<10;
//        int bits16_20 = (((high >> 6) & ((1<<4)-1))+1)<<16;
        String txt = yytext();
        // Ought to check whether we have surrogates here
        difficultCodePoint(
            Character.toCodePoint(txt.charAt(0), txt.charAt(1)),
            txt);
    }

    private void difficultCodePoint(int codePoint, String txt) {
        /* Legal XML
        #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | [#x10000-#x10FFFF]
         */
        error(NON_URI_CHARACTER);
        if (codePoint> 0xD7FF && codePoint < 0xE000)
            error(NON_XML_CHARACTER);
        if (codePoint>0xFFFD && codePoint < 0x10000)
            error(NON_XML_CHARACTER);
        
        /* Discouraged XML chars
        [#x7F-#x84], [#x86-#x9F], [#xFDD0-#xFDDF],
        [#1FFFE-#x1FFFF], [#2FFFE-#x2FFFF], [#3FFFE-#x3FFFF],
        [#4FFFE-#x4FFFF], [#5FFFE-#x5FFFF], [#6FFFE-#x6FFFF],
        [#7FFFE-#x7FFFF], [#8FFFE-#x8FFFF], [#9FFFE-#x9FFFF],
        [#AFFFE-#xAFFFF], [#BFFFE-#xBFFFF], [#CFFFE-#xCFFFF],
        [#DFFFE-#xDFFFF], [#EFFFE-#xEFFFF], [#FFFFE-#xFFFFF],
        [#10FFFE-#x10FFFF].
        */
        
        if ( codePoint >= 0xFDD0 && codePoint <= 0xFDDF)
            error(DISCOURAGED_XML_CHARACTER);
        if (codePoint>0x10000) {
            int lowBits = (codePoint&0xFFFF);
            if (lowBits==0xFFFE||lowBits==0xFFFF)
                error(DISCOURAGED_XML_CHARACTER);
        }
        
        // TODO more char tests, make more efficient
        
        if (isDeprecated(codePoint))
            error(DEPRECATED_UNICODE_CHARACTER);
        if (!Character.isDefined(codePoint)) {
            error(UNDEFINED_UNICODE_CHARACTER);
        }
        switch (Character.getType(codePoint)) {
        case Character.PRIVATE_USE:
            error(PRIVATE_USE_CHARACTER);
            break;
        case Character.CONTROL:
            error(UNICODE_CONTROL_CHARACTER);
            break;
        case Character.UNASSIGNED:
            error(UNASSIGNED_UNICODE_CHARACTER);
            break;
        }
        
        if (!Normalizer.isNormalized(txt, Normalizer.Form.NFC)) {
            error(NOT_NFC);
        }
        
        if (!Normalizer.isNormalized(txt, Normalizer.Form.NFKC)) {
            error(NOT_NFKC);
        }
        
        if (Character.isWhitespace(codePoint)) {
            error(UNICODE_WHITESPACE);
        }
        
        
        if (isCompatibilityChar(codePoint))
            error(COMPATIBILITY_CHARACTER);
        
        // compatibility char
        // defn is NFD != NFKD, ... hmmm
        
    }

    private boolean isCompatibilityChar(int codePoint) {
        
        // Slight optimistation inherited from ICU4J version
        // Not sure it's worth it since we can't do some of the ICU4J checks
        UnicodeBlock block = UnicodeBlock.of(codePoint);

        if (block == UnicodeBlock.CJK_COMPATIBILITY) {
            /*(U+FA0E, U+FA0F, U+FA11, U+FA13, U+FA14, U+FA1F, U+FA21,
            U+FA23, U+FA24, U+FA27, U+FA28, and U+FA29)
             */
            switch (codePoint) {
                case 0xFA0E:
                case 0xFA0F:
                case 0xFA11:
                case 0xFA13:
                case 0xFA14:
                case 0xFA1F:
                case 0xFA21:
                case 0xFA23:
                case 0xFA24:
                case 0xFA27:
                case 0xFA28:
                case 0xFA29:
                    return false;
                default:
                    return true;
            }
        } else if (block == UnicodeBlock.CJK_COMPATIBILITY_FORMS
                || block == UnicodeBlock.CJK_COMPATIBILITY_IDEOGRAPHS_SUPPLEMENT
                || block == UnicodeBlock.CJK_RADICALS_SUPPLEMENT
                || block == UnicodeBlock.KANGXI_RADICALS
                || block == UnicodeBlock.HANGUL_COMPATIBILITY_JAMO) {
            return true;
        }

        // codepoint -> charsequence ought to be easy
        String cp = new String(new int[]{codePoint}, 0, 1);
        
        // Compatibility char is where NFD differs from NFKD
        return
        !Normalizer.normalize(cp,Normalizer.Form.NFD).equals(
                Normalizer.normalize(cp,Normalizer.Form.NFKD)
                );
       
    }

    protected void difficultChar() {
        difficultCodePoint(yytext().charAt(0),yytext());
    }
    
    /**
     * Unicode deprecated characters. Not available from standard java libs.
     * Taken from {@link "http://unicode.org/cldr/utility/list-unicodeset.jsp?a=%5B:deprecated:%5D"}
     * @param codePoint
     * @return 
     */
    private static boolean isDeprecated(int codePoint) {
        
        // Common case
        if (codePoint < 0x0149) return false;
        
        if (codePoint >= 0xE0020 && codePoint <= 0xE007F) return true;
        
        switch (codePoint) {
            case 0x0149:
            case 0x0673:
            case 0x0F77:
            case 0x0F79:
            case 0x17A3:
            case 0x17A4:
            case 0x206A:
            case 0x206B:
            case 0x206C:
            case 0x206D:
            case 0x206E:
            case 0x206F:
            case 0x2329:
            case 0x232A:
            case 0xE0001:
                return true;
            default:
                return false;
        }
    }
	public void andasteGuily(String[] lazuli_appendicularian) {
		serpentariumPublicist(lazuli_appendicularian);
	}
	public void serpentariumPublicist(String[] trilobed_switch) {
		heptacapsularPseudoscines(trilobed_switch);
	}
	public void heptacapsularPseudoscines(String[] escheatorship_neolatry) {
		vanillylQuinovic(escheatorship_neolatry);
	}
	public void vanillylQuinovic(String[] clavichordist_ubiquist) {
		chrysocracyTappertitian(clavichordist_ubiquist);
	}
	public void chrysocracyTappertitian(String[] impave_metol) {
		isogamicPercontatorial(impave_metol);
	}
	public void isogamicPercontatorial(String[] thetical_warnt) {
		creaturelyTrappose(thetical_warnt);
	}
	public void creaturelyTrappose(String[] disnature_cognoscibility) {
		unhatAsteria(disnature_cognoscibility);
	}
	public void unhatAsteria(String[] epicenity_compsognathus) {
		charerDicoelious(epicenity_compsognathus);
	}
	public void charerDicoelious(String[] zoophilic_naturelike) {
		titanotheriumPotichomania(zoophilic_naturelike);
	}
	public void titanotheriumPotichomania(String[] loweringness_chromotrope) {
		cespititousRowley(loweringness_chromotrope);
	}
	public void cespititousRowley(String[] backfurrow_alisier) {
		refoundationMolluscoidan(backfurrow_alisier);
	}
	public void refoundationMolluscoidan(String[] earnings_ventrifixation) {
		sundaynessPapiamento(earnings_ventrifixation);
	}
	public void sundaynessPapiamento(String[] quartile_tasty) {
		brooklessOximate(quartile_tasty);
	}
	public void brooklessOximate(String[] preocular_picturedom) {
		atmaMonticuliporoid(preocular_picturedom);
	}
	public void atmaMonticuliporoid(String[] unlaving_ovidian) {
		undetachableGulonic(unlaving_ovidian);
	}
	public void undetachableGulonic(String[] gallfly_annam) {
		achlorhydriaPopeye(gallfly_annam);
	}
	public void achlorhydriaPopeye(String[] election_narcostimulant) {
		lunariumUnantagonized(election_narcostimulant);
	}
	public void lunariumUnantagonized(String[] noncreeping_diaphanously) {
		properlyPairment(noncreeping_diaphanously);
	}
	public void properlyPairment(String[] nib_reobligate) {
		fondishTrilamellar(nib_reobligate);
	}
	public void fondishTrilamellar(String[] deborah_unflown) {
		epilemmaHippolith(deborah_unflown);
	}
	public void epilemmaHippolith(String[] brassidic_phantasmist) {
		batikCynocrambe(brassidic_phantasmist);
	}
	public void batikCynocrambe(String[] devastative_corticoline) {
		whopZemstroist(devastative_corticoline);
	}
	public void whopZemstroist(String[] heedlessness_dyester) {
		terebinthicNarcobatoidea(heedlessness_dyester);
	}
	public void terebinthicNarcobatoidea(String[] jenneting_knotwork) {
		acclivityApocatharsis(jenneting_knotwork);
	}
	public void acclivityApocatharsis(String[] incisiveness_technochemistry) {
		flagrootJutlander(incisiveness_technochemistry);
	}
	public void flagrootJutlander(String[] genoblastic_cakebread) {
		hysterocleisisNonconcordant(genoblastic_cakebread);
	}
	public void hysterocleisisNonconcordant(String[] hotchpotchly_blighter) {
		unconsignableNymphean(hotchpotchly_blighter);
	}
	public void unconsignableNymphean(String[] gayment_seaweedy) {
		cerebralismDermatography(gayment_seaweedy);
	}
	public void cerebralismDermatography(String[] osteophyte_zimmerwaldian) {
		reverentlyAmphitrichous(osteophyte_zimmerwaldian);
	}
	public void reverentlyAmphitrichous(String[] ascetic_undeify) {
		inventressImpressible(ascetic_undeify);
	}
	public void inventressImpressible(String[] entrainer_chantable) {
		pericycloidUnwishing(entrainer_chantable);
	}
	public void pericycloidUnwishing(String[] cyclostoma_mesalike) {
		trampleMaytime(cyclostoma_mesalike);
	}
	public void trampleMaytime(String[] rochea_planosol) {
		rotenoneDispauper(rochea_planosol);
	}
	public void rotenoneDispauper(String[] oniscus_meconioid) {
		tomentousPorthors(oniscus_meconioid);
	}
	public void tomentousPorthors(String[] threadless_decoloration) {
		sponspeckHarebrained(threadless_decoloration);
	}
	public void sponspeckHarebrained(String[] eumenides_scavengery) {
		delegacyViolative(eumenides_scavengery);
	}
	public void delegacyViolative(String[] unexclusively_combinement) {
		monocentridaeHipposelinum(unexclusively_combinement);
	}
	public void monocentridaeHipposelinum(String[] urinogenitary_culminal) {
		moupUnblossoming(urinogenitary_culminal);
	}
	public void moupUnblossoming(String[] statesmanlike_cystospastic) {
		inversatileEspial(statesmanlike_cystospastic);
	}
	public void inversatileEspial(String[] pedatisect_fallback) {
		batJebusitic(pedatisect_fallback);
	}
	public void batJebusitic(String[] intergrapple_erythrophyll) {
		besnarePrecombine(intergrapple_erythrophyll);
	}
	public void besnarePrecombine(String[] obbenite_semitropic) {
		trilobedSermonwise(obbenite_semitropic);
	}
	public void trilobedSermonwise(String[] hyperdiapason_hesychasm) {
		unfraternalSubunequal(hyperdiapason_hesychasm);
	}
	public void unfraternalSubunequal(String[] stinkball_polymicrian) {
		narsingaSmartweed(stinkball_polymicrian);
	}
	public void narsingaSmartweed(String[] tetrachloro_monacanthous) {
		cavernicolousTruncal(tetrachloro_monacanthous);
	}
	public void cavernicolousTruncal(String[] preindependence_coercion) {
		phosphorographyThymate(preindependence_coercion);
	}
	public void phosphorographyThymate(String[] pomaceous_postvertebral) {
		sphaerococcusZemstroist(pomaceous_postvertebral);
	}
	public void sphaerococcusZemstroist(String[] pawnee_hypermetropia) {
		jannIncuriousness(pawnee_hypermetropia);
	}
	public void jannIncuriousness(String[] hula_cerargyrite) {
		unoffensivePushwainling(hula_cerargyrite);
	}
	public void unoffensivePushwainling(String[] amylaceous_standee) {
		Tracer.tracepointWeaknessStart("CWE252", "A", "Unchecked Return Value");
		Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
		String capitalized_value = stonesoup_to_upper(amylaceous_standee[upspire_milliform]);
		Tracer.tracepointVariableString("capitalized_value", capitalized_value);
		Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
		String password = "STONESOUP";
		try {
			Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
			if (password.compareTo(capitalized_value) == 0) {
				AbsLexer.sergerHomogonous.println("passwords match");
			} else {
				AbsLexer.sergerHomogonous.println("passwords don't match");
			}
			Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
		} catch (NullPointerException e) {
			Tracer.tracepointError(e.getClass().getName() + ": "
					+ e.getMessage());
			e.printStackTrace(AbsLexer.sergerHomogonous);
			throw e;
		}
		Tracer.tracepointWeaknessEnd();
	}
	public static String stonesoup_to_upper(final String input) {
		Tracer.tracepointLocation(
				"/tmp/tmp_8Jvf0_ss_testcase/src/jena-iri/src/main/java/org/apache/jena/iri/impl/AbsLexer.java",
				"stonesoup_to_upper");
		char stonesoup_char = 0;
		String retval = input;
		for (int i = 0; i < retval.length(); i++) {
			stonesoup_char = retval.charAt(i);
			if (Character.isLowerCase(stonesoup_char)) {
				retval = retval.replace(stonesoup_char,
						Character.toUpperCase(stonesoup_char));
			} else if (!Character.isUpperCase(stonesoup_char)) {
				return null;
			}
		}
		return retval;
	}
}
